import json
from optparse import OptionParser

from gurke.stream.gherkin_events import GherkinEvents
from gurke.stream.source_events import SourceEvents

parser = OptionParser()
parser.add_option(
    "--no-source", action="store_false", dest="print_source", default=True, help="don't print source events"
)
parser.add_option("--no-ast", action="store_false", dest="print_ast", default=True, help="don't print ast events")
parser.add_option(
    "--no-pickles", action="store_false", dest="print_pickles", default=True, help="don't print pickle events"
)

(options, args) = parser.parse_args()

source_events = SourceEvents(args)
gherkin_events = GherkinEvents(options)

for source_event in source_events.enum():
    for event in gherkin_events.enum(source_event):
        print(json.dumps(event))



from ..ast_builder import AstBuilder
from ..errors import CompositeParserException, ParserError
from ..parser import Parser
from ..pickles.compiler import Compiler
from ..stream.id_generator import IdGenerator


def create_errors(errors, uri):
    for error in errors:
        yield {
            "parseError": {
                "source": {"uri": uri, "location": error.location},
                "message": str(error),
            }
        }


class GherkinEvents:
    def __init__(self, options):
        self.options = options
        self.id_generator = IdGenerator()
        self.parser = Parser(ast_builder=AstBuilder(self.id_generator))
        self.compiler = Compiler(self.id_generator)

    def enum(self, source_event):
        uri = source_event["source"]["uri"]
        source = source_event["source"]["data"]

        try:
            gherkin_document = self.parser.parse(source)
            gherkin_document["uri"] = uri

            if self.options.print_source:
                yield source_event

            if self.options.print_ast:
                yield {"gherkinDocument": gherkin_document}

            if self.options.print_pickles:
                pickles = self.compiler.compile(gherkin_document)
                for pickle in pickles:
                    yield {"pickle": pickle}
        except CompositeParserException as e:
            for event in create_errors(e.errors, uri):
                yield event
        except ParserError as e:
            for event in create_errors([e], uri):
                yield event



import re
from collections import defaultdict
from typing import Any

from .dialect import Dialect
from .errors import NoSuchLanguageException
from .location import Location
from .token import Token

KEYWORD_PREFIX_BULLET = "^(\\s*[*+-]\\s*)"
KEYWORD_PREFIX_HEADER = "^(#{1,6}\\s)"


class GherkinInMarkdownTokenMatcher:
    LANGUAGE_RE = re.compile(r"^\s*#\s*language\s*:\s*([a-zA-Z\-_]+)\s*$")

    def __init__(self, dialect_name: str = "en") -> None:
        self._default_dialect_name = dialect_name
        self._change_dialect(dialect_name)
        self.reset()

    def reset(self) -> None:
        if self.dialect_name != self._default_dialect_name:
            self._change_dialect(self._default_dialect_name)
        self._indent_to_remove = 0
        self._active_doc_string_separator = None
        self.matched_feature_line = False

    def match_FeatureLine(self, token: Token) -> bool:  # noqa: N802
        if self.matched_feature_line:
            self._set_token_matched(token, None)
            return False

        # We first try to match "# Feature: blah"
        result = self._match_title_line(KEYWORD_PREFIX_HEADER, self.dialect.feature_keywords, ":", token, "FeatureLine")
        # If we didn't match "# Feature: blah", we still match this line
        # as a FeatureLine.
        # The reason for this is that users may not want to be constrained by having this as their fist line.

        if not result:
            self._set_token_matched(token, "FeatureLine", token.line.get_line_text())
            result = True
        self.matched_feature_line = result
        return result

    def match_RuleLine(self, token):  # noqa: N802
        return self._match_title_line(KEYWORD_PREFIX_HEADER, self.dialect.rule_keywords, ":", token, "RuleLine")

    def match_ScenarioLine(self, token):  # noqa: N802
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER, self.dialect.scenario_keywords, ":", token, "ScenarioLine"
        ) or self._match_title_line(
            KEYWORD_PREFIX_HEADER, self.dialect.scenario_outline_keywords, ":", token, "ScenarioLine"
        )

    def match_BackgroundLine(self, token):  # noqa: N802
        return self._match_title_line(
            KEYWORD_PREFIX_HEADER, self.dialect.background_keywords, ":", token, "BackgroundLine"
        )

    def match_ExamplesLine(self, token):  # noqa: N802
        return self._match_title_line(KEYWORD_PREFIX_HEADER, self.dialect.examples_keywords, ":", token, "ExamplesLine")

    def match_TableRow(self, token: Token) -> bool:  # noqa: N802
        # Gherkin tables must be indented 2-5 spaces in order to be distinguidedn from non-Gherkin tables

        if re.match("^\\s?\\s?\\s?\\s?\\s?\\|", token.line.get_line_text(0)):
            table_cells = token.line.table_cells
            if self._is_gfm_table_separator(table_cells):
                return False

            self._set_token_matched(token, "TableRow", keyword="|", items=token.line.table_cells)

            return True
        return False

    def _is_gfm_table_separator(self, table_cells):
        text_of_table_cells = map(lambda x: x["text"], table_cells)
        separator_values = list(filter(lambda x: re.match("^:?-+:?$", x), text_of_table_cells))
        return len(separator_values) > 0

    def match_StepLine(self, token):  # noqa: N802
        non_star_step_keywords = (
            self.dialect.given_keywords
            + self.dialect.when_keywords
            + self.dialect.then_keywords
            + self.dialect.and_keywords
            + self.dialect.but_keywords
        )
        return self._match_title_line(KEYWORD_PREFIX_BULLET, non_star_step_keywords, "", token, "StepLine")

    def match_Comment(self, token):  # noqa: N802
        result = False
        if token.line.startswith("|"):
            table_cells = token.line.table_cells
            if self._is_gfm_table_separator(table_cells):
                result = True
        self._set_token_matched(token, "Comment")
        return result

    def match_Empty(self, token):  # noqa: N802
        result = False
        if token.line.is_empty():
            result = True
        if (
            not self.match_TagLine(token)
            and not self.match_FeatureLine(token)
            and not self.match_ScenarioLine(token)
            and not self.match_BackgroundLine(token)
            and not self.match_ExamplesLine(token)
            and not self.match_RuleLine(token)
            and not self.match_TableRow(token)
            and not self.match_Comment(token)
            and not self.match_Language(token)
            and not self.match_DocStringSeparator(token)
            and not self.match_EOF(token)
            and not self.match_StepLine(token)
        ):
            # neutered
            result = True

        if result:
            self._set_token_matched(token, "Empty", indent=0)
            return result
        return False

    # We've made a deliberate choice not to support `# language: [ISO 639-1]` headers or similar
    # in Markdown. Users should specify a language globally.
    def match_Language(self, token):  # noqa: N802
        if not token:
            raise ValueError("no token")
        return False

    def match_TagLine(self, token):  # noqa: N802
        tags = []
        matching_tags = re.finditer("`(@[^`]+)`", token.line.get_line_text())
        idx = 0
        for match in matching_tags:
            tags.append({"column": token.line.indent + match.start(idx) + 2, "text": match.group(1)})

        if len(tags) == 0:
            return False

        self._set_token_matched(token, "TagLine", items=tags)
        return True

    def match_DocStringSeparator(self, token):  # noqa: N802
        if not self._active_doc_string_separator:
            # open
            return (
                self._match_DocStringSeparator(token, '"""', True)
                or self._match_DocStringSeparator(token, "````", True)
                or self._match_DocStringSeparator(token, "```", True)
            )

        # close
        return self._match_DocStringSeparator(token, self._active_doc_string_separator, False)

    def _match_DocStringSeparator(self, token, separator, is_open):  # noqa: N802
        if not token.line.startswith(separator):
            return False

        content_type = ""
        if is_open:
            content_type = token.line.get_rest_trimmed(len(separator))
            self._active_doc_string_separator = separator
            self._indent_to_remove = token.line.indent
        else:
            self._active_doc_string_separator = None
            self._indent_to_remove = 0

        # TODO: Use the separator as keyword. That's needed for pretty printing.
        self._set_token_matched(token, "DocStringSeparator", content_type, separator)
        return True

    def match_Other(self, token):  # noqa: N802
        # take the entire line, except removing DocString indents
        text = token.line.get_line_text(self._indent_to_remove)
        self._set_token_matched(token, "Other", self._unescaped_docstring(text), indent=0)
        return True

    def match_EOF(self, token):  # noqa: N802
        if not token.eof():
            return False

        self._set_token_matched(token, "EOF")
        return True

    def _match_title_line(self, prefix: Any, keywords: Any, keyword_suffix: Any, token: Token, token_type: Any) -> bool:
        keywords_or_list = "|".join(map(lambda x: re.escape(x), keywords))
        match = re.search("{}({}){}(.*)".format(prefix, keywords_or_list, keyword_suffix), token.line.get_line_text())
        indent = token.line.indent
        matched_keyword_type = None
        if match:
            matched_keyword = match.group(2)
            indent += len(match.group(1))

            # only set the keyword type if this is a step keyword
            if matched_keyword in self.keyword_types:
                matched_keyword_type = self.keyword_types[matched_keyword][0]

            self._set_token_matched(
                token,
                token_type,
                match.group(3).strip(),
                matched_keyword,
                keyword_type=matched_keyword_type,
                indent=indent,
            )
            return True
        return False

    def _set_token_matched2(
        self,
        token,
        matched,
        indent=None,
    ):
        token.matched_gherkin_dialect = self.dialect_name
        if indent is not None:
            token.matched_indent = indent
        else:
            token.matched_indent = token.line.indent if token.line else 0
        token.location.column = token.matched_indent + 1
        return matched

    def _set_token_matched(
        self,
        token: Token,
        matched_type: Any,
        text: Any = None,
        keyword: Any = None,
        keyword_type: Any = None,
        indent: Any = None,
        items: Any = None,
    ) -> None:
        if items is None:
            items = []
        token.matched_type = matched_type
        # text == '' should not result in None
        token.matched_text = text.rstrip("\r\n") if text is not None else None
        token.matched_keyword = keyword
        token.matched_keyword_type = keyword_type
        if indent is not None:
            token.matched_indent = indent
        else:
            token.matched_indent = token.line.indent if token.line else 0
        token.matched_items = items
        token.location.column = token.matched_indent + 1
        token.matched_gherkin_dialect = self.dialect_name

    def _change_dialect(self, dialect_name: str, location: Location = None) -> None:
        dialect = Dialect.for_name(dialect_name)
        if not dialect:
            raise NoSuchLanguageException(dialect_name, location)

        self.dialect_name = dialect_name
        self.dialect = dialect
        self.keyword_types = defaultdict(list)
        for keyword in self.dialect.given_keywords:
            self.keyword_types[keyword].append("Context")
        for keyword in self.dialect.when_keywords:
            self.keyword_types[keyword].append("Action")
        for keyword in self.dialect.then_keywords:
            self.keyword_types[keyword].append("Outcome")
        for keyword in self.dialect.and_keywords + self.dialect.but_keywords:
            self.keyword_types[keyword].append("Conjunction")

    def _unescaped_docstring(self, text):
        if self._active_doc_string_separator == '"""':
            return text.replace('\\"\\"\\"', '"""')
        if self._active_doc_string_separator == "```":
            return text.replace("\\`\\`\\`", "```")
        return text



from __future__ import annotations

import io
import json
import os
from typing import Dict, List, Optional

DIALECT_FILE_PATH = os.path.join(os.path.dirname(__file__), "gherkin-languages.json")

with io.open(DIALECT_FILE_PATH, "r", encoding="utf-8") as file:
    DIALECTS = json.load(file)


class Dialect(object):
    @classmethod
    def for_name(cls, name: str) -> Optional[Dialect]:
        return cls(DIALECTS[name]) if name in DIALECTS else None

    def __init__(self, spec: Dict[str, List[str]]) -> None:
        self.spec = spec

    @property
    def feature_keywords(self) -> List[str]:
        return self.spec["feature"]

    @property
    def rule_keywords(self) -> List[str]:
        return self.spec["rule"]

    @property
    def scenario_keywords(self) -> List[str]:
        return self.spec["scenario"]

    @property
    def scenario_outline_keywords(self) -> List[str]:
        return self.spec["scenarioOutline"]

    @property
    def background_keywords(self) -> List[str]:
        return self.spec["background"]

    @property
    def examples_keywords(self) -> List[str]:
        return self.spec["examples"]

    @property
    def given_keywords(self) -> List[str]:
        return self.spec["given"]

    @property
    def when_keywords(self) -> List[str]:
        return self.spec["when"]

    @property
    def then_keywords(self) -> List[str]:
        return self.spec["then"]

    @property
    def and_keywords(self) -> List[str]:
        return self.spec["and"]

    @property
    def but_keywords(self) -> List[str]:
        return self.spec["but"]



import io
import os
from typing import Union

from .gherkin_line import GherkinLine
from .location import Location
from .token import Token


class TokenScanner(object):
    def __init__(self, path_or_str: Union[os.PathLike[str], str]) -> None:
        if isinstance(path_or_str, os.PathLike):
            self.io = io.open(path_or_str, "r", encoding="utf8")
        else:
            self.io = io.StringIO(path_or_str)
        self.line_number = 0

    def read(self) -> Token:
        self.line_number += 1
        line = self.io.readline()
        return Token(GherkinLine(line, self.line_number), Location(self.line_number, None))

    def __del__(self) -> None:
        try:
            self.io.close()
        except AttributeError:
            pass



import re

from ..stream.id_generator import IdGenerator


class Compiler(object):
    def __init__(self, id_generator=None):
        self.id_generator = id_generator
        if self.id_generator is None:
            self.id_generator = IdGenerator()

    def compile(self, gherkin_document):
        pickles = []
        if "feature" not in gherkin_document:
            return pickles

        feature = gherkin_document["feature"]
        if not feature["children"]:
            return pickles

        uri = gherkin_document["uri"]
        feature_tags = feature["tags"]
        language = feature["language"]
        background_steps = []
        for scenario_definition in feature["children"]:
            if "background" in scenario_definition:
                if scenario_definition["background"]["steps"]:
                    background_steps += scenario_definition["background"]["steps"]
            elif "rule" in scenario_definition:
                self._compile_rule(uri, feature_tags, background_steps, scenario_definition["rule"], language, pickles)
            else:
                scenario = scenario_definition["scenario"]
                args = (uri, feature_tags, background_steps, scenario, language, pickles)
                if not scenario["examples"]:
                    self._compile_scenario(*args)
                else:
                    self._compile_scenario_outline(*args)
        return pickles

    def _compile_rule(self, uri, feature_tags, feature_background_steps, rule, language, pickles):
        tags = list(feature_tags) + list(rule["tags"])
        background_steps = []
        background_steps += feature_background_steps
        for scenario_definition in rule["children"]:
            if "background" in scenario_definition:
                if scenario_definition["background"]["steps"]:
                    background_steps += scenario_definition["background"]["steps"]
            else:
                scenario = scenario_definition["scenario"]
                args = (uri, tags, background_steps, scenario, language, pickles)
                if not scenario["examples"]:
                    self._compile_scenario(*args)
                else:
                    self._compile_scenario_outline(*args)
        return pickles

    def _compile_scenario(self, uri, inherited_tags, background_steps, scenario, language, pickles):
        tags = list(inherited_tags) + list(scenario["tags"])
        last_keyword_type = "Unknown"
        steps = list()
        if scenario["steps"]:
            for step in background_steps + scenario["steps"]:
                last_keyword_type = last_keyword_type if step["keywordType"] == "Conjunction" else step["keywordType"]
                steps.append(self._pickle_step(step, last_keyword_type))

        pickle = {
            "astNodeIds": [scenario["id"]],
            "id": self.id_generator.get_next_id(),
            "tags": self._pickle_tags(tags),
            "name": scenario["name"],
            "language": language,
            "steps": steps,
            "uri": uri,
        }
        pickles.append(pickle)

    def _compile_scenario_outline(self, uri, inherited_tags, background_steps, scenario, language, pickles):
        for examples in (e for e in scenario["examples"] if "tableHeader" in e):
            variable_cells = examples["tableHeader"]["cells"]

            for values in examples["tableBody"]:
                value_cells = values["cells"]
                tags = list(inherited_tags) + list(scenario["tags"]) + list(examples["tags"])
                last_keyword_type = None
                steps = list()
                if scenario["steps"]:
                    for step in background_steps:
                        last_keyword_type = (
                            last_keyword_type if step["keywordType"] == "Conjunction" else step["keywordType"]
                        )
                        steps.append(self._pickle_step(step, last_keyword_type))

                if scenario["steps"]:
                    for outline_step in scenario["steps"]:
                        last_keyword_type = (
                            last_keyword_type
                            if outline_step["keywordType"] == "Conjunction"
                            else outline_step["keywordType"]
                        )
                        step_text = self._interpolate(outline_step["text"], variable_cells, value_cells)
                        argument = self._create_pickle_arguments(outline_step, variable_cells, value_cells)
                        _pickle_step = {
                            "astNodeIds": [outline_step["id"], values["id"]],
                            "id": self.id_generator.get_next_id(),
                            "type": last_keyword_type,
                            "text": step_text,
                        }
                        if argument is not None:
                            _pickle_step["argument"] = argument

                        steps.append(_pickle_step)

                pickle = {
                    "astNodeIds": [scenario["id"], values["id"]],
                    "id": self.id_generator.get_next_id(),
                    "name": self._interpolate(scenario["name"], variable_cells, value_cells),
                    "language": language,
                    "steps": steps,
                    "tags": self._pickle_tags(tags),
                    "uri": uri,
                }
                pickles.append(pickle)

    def _create_pickle_arguments(self, step, variables, values):
        if "dataTable" in step:
            table = {"rows": []}
            for row in step["dataTable"]["rows"]:
                cells = [{"value": self._interpolate(cell["value"], variables, values)} for cell in row["cells"]]
                table["rows"].append({"cells": cells})
            return {"dataTable": table}

        if "docString" in step:
            argument = step["docString"]
            docstring = {"content": self._interpolate(argument["content"], variables, values)}
            if "mediaType" in argument:
                docstring["mediaType"] = self._interpolate(argument["mediaType"], variables, values)
            return {"docString": docstring}

        return None

    def _interpolate(self, name, variable_cells, value_cells):
        if name is None:
            return name

        for n, variable_cell in enumerate(variable_cells):
            value_cell = value_cells[n]
            # For the case of trailing backslash, re-escaping backslashes are needed
            reescaped_value = re.sub(r"\\", r"\\\\", value_cell["value"])
            name = re.sub("<{0[value]}>".format(variable_cell), reescaped_value, name)
        return name

    def _pickle_step(self, step, keyword_type):
        pickle_step = {
            "astNodeIds": [step["id"]],
            "id": self.id_generator.get_next_id(),
            "type": keyword_type,
            "text": step["text"],
        }
        argument = self._create_pickle_arguments(step, [], [])
        if argument is not None:
            pickle_step["argument"] = argument
        return pickle_step

    def _pickle_tags(self, tags):
        return [self._pickle_tag(tag) for tag in tags]

    def _pickle_tag(self, tag):
        return {"astNodeId": tag["id"], "name": tag["name"]}



import io


def source_event(path):
    return {
        "source": {
            "uri": path,
            "data": io.open(path, "r", encoding="utf8", newline="").read(),
            "mediaType": "text/x.cucumber.gherkin+plain",
        }
    }


class SourceEvents:
    def __init__(self, paths):
        self.paths = paths

    def enum(self):
        return map(source_event, self.paths)
